@@ -4,7 +4,7 @@ use 5.010;
 use strict;
 use warnings;
 
-our $VERSION = '0.09';
+our $VERSION = '0.10';
 
 use Cassandra::Client::AsyncAnyEvent;
 use Cassandra::Client::AsyncEV;
@@ -16,6 +16,7 @@ use Cassandra::Client::Policy::Retry::Default;
 use Cassandra::Client::Policy::Retry;
 use Cassandra::Client::Policy::Throttle::Adaptive;
 use Cassandra::Client::Pool;
+use Cassandra::Client::TLSHandling;
 use Cassandra::Client::Util qw/series whilst/;
 
 use Clone qw/clone/;
@@ -30,7 +31,6 @@ sub new {
         connected         => 0,
         connect_callbacks => undef,
         shutdown          => 0,
-        shutdown_cb       => undef,
 
         active_queries    => 0,
     }, $class;
@@ -49,11 +49,13 @@ sub new {
         client   => $self,
         options  => $options,
         metadata => $metadata,
+        async_io => $async_io,
     );
     my $command_queue= Cassandra::Client::Policy::Queue::Default->new(
         %{ $options->{command_queue_config} || {} },
     );
     my $retry_policy= Cassandra::Client::Policy::Retry::Default->new();
+    my $tls= $options->{tls} ? Cassandra::Client::TLSHandling->new() : undef;
 
     $self->{options}= $options;
     $self->{async_io}= $async_io;
@@ -61,6 +63,7 @@ sub new {
     $self->{pool}= $pool;
     $self->{command_queue}= $command_queue;
     $self->{retry_policy}= $retry_policy;
+    $self->{tls}= $tls;
     if ($options->{throttler}) {
         $options->{throttler}= 'Adaptive' if $options->{throttler} eq 'AdaptiveThrottler'; # Temporary.
         my $throttler_class= "Cassandra::Client::Policy::Throttle::$options->{throttler}";
@@ -109,9 +112,7 @@ sub _connect {
             },
             sub {
                 my ($next)= @_;
-                $self->{pool}{datacenter}= $connection->{datacenter}; #XXX
-                $self->{pool}->add($connection);
-                $self->{pool}->warmup($next);
+                $self->{pool}->init($next, $connection);
             },
         ], sub {
             my $error= shift;
@@ -131,30 +132,14 @@ sub _connect {
     return;
 }
 
-sub _shutdown {
-    my ($self, $callback)= @_;
-
-    if ($self->{shutdown_cb}) {
-        push @{$self->{shutdown_cb}}, $callback;
-        return;
-    }
-    if ($self->{shutdown} && !$self->{shutdown_cb}) {
-        return _cb($callback);
-    }
-    if (!$self->{connected}) {
-        return _cb($callback, 'Cannot shutdown: not connected');
-    }
+sub shutdown {
+    my ($self)= @_;
 
+    return if $self->{shutdown};
     $self->{shutdown}= 1;
     $self->{connected}= 0;
-    $self->{shutdown_cb}= [$callback];
 
-    $self->{pool}->shutdown(sub {
-        my ($error)= @_;
-        my $cb= $self->{shutdown_cb};
-        $self->{shutdown_cb}= undef;
-        _cb($_, $error) for @$cb;
-    });
+    $self->{pool}->shutdown;
 
     return;
 }
@@ -205,13 +190,21 @@ sub _prepare {
 
 sub _execute {
     my ($self, $callback, $query, $params, $attribs)= @_;
-    $self->_command("execute_prepared", $callback, [ \$query, clone($params), clone($attribs) ]);
+
+    my $attribs_clone= clone($attribs);
+    $attribs_clone->{consistency} ||= $self->{options}{default_consistency};
+
+    $self->_command("execute_prepared", $callback, [ \$query, clone($params), $attribs_clone ]);
     return;
 }
 
 sub _batch {
     my ($self, $callback, $queries, $attribs)= @_;
-    $self->_command("execute_batch", $callback, [ clone($queries), clone($attribs) ]);
+
+    my $attribs_clone= clone($attribs);
+    $attribs_clone->{consistency} ||= $self->{options}{default_consistency};
+
+    $self->_command("execute_batch", $callback, [ clone($queries), $attribs_clone ]);
     return;
 }
 
@@ -326,6 +319,8 @@ sub _command_failed {
     my $retry_decision;
     if ($error->{do_retry}) {
         $retry_decision= Cassandra::Client::Policy::Retry::retry;
+    } elsif ($error->{request_error}) {
+        $retry_decision= $self->{retry_policy}->on_request_error(undef, undef, $error, ($command_info->{retries}||0));
     } elsif ($error->code == 0x1100) {
         $retry_decision= $self->{retry_policy}->on_write_timeout(undef, @$error{qw/cl write_type blockfor received/}, ($command_info->{retries}||0));
     } elsif ($error->code == 0x1200) {
@@ -396,6 +391,16 @@ sub _each_page {
     return;
 }
 
+sub DESTROY {
+    local $@;
+    return if ${^GLOBAL_PHASE} eq 'DESTRUCT';
+
+    my $self= shift;
+    if ($self->{connected}) {
+        $self->shutdown;
+    }
+}
+
 
 # Utility functions for callers
 sub _get_stacktrace {
@@ -496,7 +501,6 @@ PUBLIC_METHODS: {
         execute
         each_page
         prepare
-        shutdown
         wait_for_schema_agreement
     /) {
         *{$_}=               _mksync        (\&{"_$_"});
@@ -581,7 +585,7 @@ Default max page size to pass to the server. This defaults to C<5000>. Note that
 
 =item max_connections
 
-Maximum amount of connections to keep open in the Cassandra connection pool. Defaults to C<2> for historical reasons, will be raised in a future version.
+Maximum amount of connections to keep open in the Cassandra connection pool. Defaults to C<2> for historical reasons, raise this if appropriate.
 
 =item timer_granularity
 
