@@ -8,6 +8,9 @@ use Ref::Util qw/is_arrayref/;
 use IO::Socket::INET;
 use Errno;
 use Socket qw/SOL_SOCKET IPPROTO_TCP SO_KEEPALIVE TCP_NODELAY/;
+use Scalar::Util qw/weaken/;
+use Net::SSLeay qw/ERROR_WANT_READ ERROR_WANT_WRITE ERROR_NONE/;
+
 use Cassandra::Client::Util;
 use Cassandra::Client::Protocol qw/
     :constants
@@ -35,7 +38,7 @@ use Cassandra::Client::Decoder qw/
 /;
 use Cassandra::Client::Error;
 use Cassandra::Client::ResultSet;
-use Scalar::Util qw/weaken/;
+use Cassandra::Client::TLSHandling;
 
 use constant STREAM_ID_LIMIT => 32768;
 
@@ -68,6 +71,9 @@ sub new {
         pending_write   => undef,
         shutdown        => 0,
         read_buffer     => \(my $empty= ''),
+
+        tls             => undef,
+        tls_want_write  => undef,
     }, $class;
     weaken($self->{async_io});
     weaken($self->{client});
@@ -80,7 +86,7 @@ sub get_local_status {
     series([
         sub {
             my ($next)= @_;
-            $self->execute_prepared($next, \"select key, data_center, host_id, broadcast_address, rack, release_version, tokens, schema_version from system.local", undef, { consistency => 'local_one' });
+            $self->execute_prepared($next, \"select key, data_center, host_id, broadcast_address, rack, release_version, tokens, schema_version from system.local");
         },
         sub {
             my ($next, $result)= @_;
@@ -109,7 +115,7 @@ sub get_peers_status {
     series([
         sub {
             my ($next)= @_;
-            $self->execute_prepared($next, \"select peer, data_center, host_id, preferred_ip, rack, release_version, tokens, schema_version from system.peers", undef, { consistency => 'local_one' });
+            $self->execute_prepared($next, \"select peer, data_center, host_id, preferred_ip, rack, release_version, tokens, schema_version from system.peers");
         },
         sub {
             my ($next, $result)= @_;
@@ -186,10 +192,9 @@ sub execute_prepared {
         };
     }
 
-    my $chosen_consistency= $attr->{consistency} || $self->{options}{default_consistency};
-    my $consistency= ($chosen_consistency ? $consistency_lookup{$chosen_consistency} : CONSISTENCY_ONE);
+    my $consistency= $consistency_lookup{$attr->{consistency} || 'one'};
     if (!defined $consistency) {
-        return $callback->("Invalid consistency level specified");
+        return $callback->("Invalid consistency level specified: $attr->{consistency}");
     }
 
     my $page_size= (0+($attr->{page_size} || $self->{options}{max_page_size} || 0)) || undef;
@@ -208,7 +213,11 @@ sub execute_prepared {
         }
 
         if ($code != OPCODE_RESULT) {
-            return $callback->("Expected a RESULT frame but got something else; considering the query failed");
+            # This shouldn't ever happen...
+            return $callback->(Cassandra::Client::Error->new(
+                message         => "Expected a RESULT frame but got something else; considering the query failed",
+                request_error   => 1,
+            ));
         }
 
         $self->decode_result($callback, $prepared, $_[2]);
@@ -280,10 +289,9 @@ sub execute_batch {
         }
     }
 
-    my $chosen_consistency= $attribs->{consistency} || $self->{options}{default_consistency};
-    my $consistency= ($chosen_consistency ? $consistency_lookup{$chosen_consistency} : CONSISTENCY_ONE);
+    my $consistency= $consistency_lookup{$attribs->{consistency} || 'one'};
     if (!defined $consistency) {
-        return $callback->("Invalid consistency level specified");
+        return $callback->("Invalid consistency level specified: $attribs->{consistency}");
     }
 
     my $batch_frame= pack('Cn', $batch_type, (0+@prepared));
@@ -304,7 +312,11 @@ sub execute_batch {
         }
 
         if ($code != OPCODE_RESULT) {
-            return $callback->("Expected a RESULT frame but got something else; considering the query failed");
+            # This shouldn't ever happen...
+            return $callback->(Cassandra::Client::Error->new(
+                message         => "Expected a RESULT frame but got something else; considering the batch failed",
+                request_error   => 1,
+            ));
         }
 
         $self->decode_result($callback, undef, $_[2]);
@@ -542,7 +554,7 @@ sub handshake {
         sub {
             my ($next)= @_;
             if ($self->{options}{keyspace}) {
-                return $self->execute_prepared($next, \('use "'.$self->{options}{keyspace}.'"'), undef, { consistency => 'local_one' });
+                return $self->execute_prepared($next, \('use "'.$self->{options}{keyspace}.'"'));
             }
             return $next->();
         },
@@ -573,8 +585,10 @@ sub handshake {
 sub authenticate {
     my ($self, $callback, $authenticator)= @_;
 
-    my $user= $self->{options}{username};
-    my $pass= $self->{options}{password};
+    my $user= "$self->{options}{username}";
+    my $pass= "$self->{options}{password}";
+    utf8::encode($user) if utf8::is_utf8($user);
+    utf8::encode($pass) if utf8::is_utf8($pass);
 
     if (!$user || !$pass) {
         return $callback->("Server expected authentication using <$authenticator> but no credentials were set");
@@ -634,12 +648,21 @@ sub connect {
     my ($self, $callback)= @_;
     return $callback->() if $self->{connected};
 
-    if ($self->{connecting}) {
-        push @{$self->{connecting}}, $callback;
+    if ($self->{connecting}++) {
+        warn "BUG: Calling connect twice?";
+        return $callback->("Internal bug: called connect twice.");
         return;
     }
 
-    $self->{connecting}= [$callback];
+    if ($self->{options}{tls}) {
+        eval {
+            $self->{tls}= $self->{client}{tls}->new_conn;
+            1;
+        } or do {
+            my $error= $@ || "unknown TLS error";
+            return $callback->($error);
+        };
+    }
 
     my $socket; {
         local $@;
@@ -652,7 +675,8 @@ sub connect {
         );
 
         unless ($socket) {
-            return $callback->("Could not connect: $@");
+            my $error= "Could not connect: $@";
+            return $callback->($error);
         }
 
         $socket->setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1);
@@ -661,31 +685,37 @@ sub connect {
 
     $self->{socket}= $socket;
     $self->{fileno}= $socket->fileno;
-    $self->setup_io;
-    $self->handshake(sub {
-        my $st= shift;
-        $self->{connected}= 1;
-        $_->($st) for @{$self->{connecting}};
-        undef $self->{connecting};
-    });
-
-    return;
-}
-
-sub setup_io {
-    my ($self)= @_;
     $self->{async_io}->register($self->{fileno}, $self);
     $self->{async_io}->register_read($self->{fileno});
 
+    # We create a fake buffer, to ensure we wait until we can actually write
     $self->{pending_write}= '';
     $self->{async_io}->register_write($self->{fileno});
+
+    if ($self->{options}{tls}) {
+        Net::SSLeay::set_fd(${$self->{tls}}, $self->{fileno});
+        Net::SSLeay::set_connect_state(${$self->{tls}});
+    }
+
+    $self->handshake(sub {
+        my $error= shift;
+        $self->{connected}= 1;
+        if ($error) {
+            $self->shutdown("Failed to connect: $error");
+        }
+        return $callback->($error);
+    });
+
     return;
 }
 
 sub request {
     # my $body= $_[3] (let's avoid copying that blob). Yes, this code assumes ownership of the body.
     my ($self, $cb, $opcode)= @_;
-    return $cb->("Connection shutting down") if $self->{shutdown};
+    return $cb->(Cassandra::Client::Error->new(
+        message => "Connection shutting down",
+        request_error => 1,
+    )) if $self->{shutdown};
 
     my $pending= $self->{pending_streams};
 
@@ -693,7 +723,10 @@ sub request {
     my $attempts= 0;
     while (exists($pending->{$stream_id}) || $stream_id >= STREAM_ID_LIMIT) {
         $stream_id= (++$stream_id) % STREAM_ID_LIMIT;
-        return $cb->("Cannot find a stream ID to post query with") if ++$attempts >= STREAM_ID_LIMIT;
+        return $cb->(Cassandra::Client::Error->new(
+            message => "Cannot find a stream ID to post query with",
+            request_error => 1,
+        )) if ++$attempts >= STREAM_ID_LIMIT;
     }
     $self->{last_stream_id}= $stream_id;
     $pending->{$stream_id}= [$cb, $self->{async_io}->deadline($self->{fileno}, $stream_id, $self->{request_timeout})];
@@ -713,28 +746,69 @@ sub request {
             last WRITE;
         }
 
-        my $length= length $data;
-        my $result= syswrite($self->{socket}, $data, $length);
-        if ($result && $result == $length) {
-            # All good
-        } elsif (defined $result || $!{EAGAIN}) {
-            substr($data, 0, $result, '') if $result;
-            $self->{pending_write}= $data;
-            $self->{async_io}->register_write($self->{fileno});
+        if ($self->{tls}) {
+            my $length= length $data;
+            my $rv= Net::SSLeay::write(${$self->{tls}}, $data);
+            if ($rv == $length) {
+                # All good
+            } elsif ($rv > 0) {
+                # Partital write
+                substr($data, 0, $rv, '');
+                $self->{pending_write}= $data;
+                $self->{async_io}->register_write($self->{fileno});
+            } else {
+                $rv= Net::SSLeay::get_error(${$self->{tls}}, $rv);
+                if ($rv == ERROR_WANT_WRITE || $rv == ERROR_WANT_READ || $rv == ERROR_NONE) {
+                    # Ok...
+                    $self->{pending_write}= $data;
+                    if ($rv == ERROR_WANT_READ) {
+                        $self->{tls_want_write}= 1;
+                    } else {
+                        $self->{async_io}->register_write($self->{fileno});
+                    }
+                } else {
+                    # We failed to send the request.
+                    my $error= Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error());
+
+                    # We never actually sent our request, so take it out again
+                    my $my_stream= delete $pending->{$stream_id};
+
+                    $self->shutdown($error);
+
+                    # Now fail our stream properly, but include the retry notice
+                    $my_stream->[0]->(Cassandra::Client::Error->new(
+                        message       => "Disconnected: $error",
+                        do_retry      => 1,
+                        request_error => 1,
+                    ));
+                }
+            }
+
         } else {
-            # Oh, we failed to send out the request. That's bad. Let's first find out what happened.
-            my $error= $!;
+            my $length= length $data;
+            my $result= syswrite($self->{socket}, $data, $length);
+            if ($result && $result == $length) {
+                # All good
+            } elsif (defined $result || $!{EAGAIN}) {
+                substr($data, 0, $result, '') if $result;
+                $self->{pending_write}= $data;
+                $self->{async_io}->register_write($self->{fileno});
+            } else {
+                # Oh, we failed to send out the request. That's bad. Let's first find out what happened.
+                my $error= $!;
 
-            # We never actually sent our request, so take it out again
-            my $my_stream= delete $pending->{$stream_id};
+                # We never actually sent our request, so take it out again
+                my $my_stream= delete $pending->{$stream_id};
 
-            $self->shutdown(undef, $error);
+                $self->shutdown($error);
 
-            # Now fail our stream properly, but include the retry notice
-            $my_stream->[0]->(Cassandra::Client::Error->new(
-                message  => "Disconnected: $error",
-                do_retry => 1,
-            ));
+                # Now fail our stream properly, but include the retry notice
+                $my_stream->[0]->(Cassandra::Client::Error->new(
+                    message       => "Disconnected: $error",
+                    do_retry      => 1,
+                    request_error => 1,
+                ));
+            }
         }
     }
 
@@ -749,19 +823,50 @@ sub can_read {
 
 READ:
     while (!$self->{shutdown}) {
-        my $read_cnt= sysread($self->{socket}, $BUFFER, 10240, $bufsize);
         my $read_something;
-        if ($read_cnt) {
-            $bufsize += $read_cnt;
-            $read_something= 1;
 
-        } elsif (!defined $read_cnt) {
-            if (!$!{EAGAIN}) {
-                my $error= "$!";
-                $shutdown_when_done= $error;
+        if ($self->{tls}) {
+            my ($bytes, $rv)= Net::SSLeay::read(${$self->{tls}});
+            if (length $bytes) {
+                $BUFFER .= $bytes;
+                $bufsize += $rv;
+                $read_something= 1;
+            }
+
+            if ($rv <= 0) {
+                $rv= Net::SSLeay::get_error(${$self->{tls}}, $rv);
+                if ($rv == ERROR_WANT_WRITE) {
+                    $self->{async_io}->register_write($self->{fileno});
+                } elsif ($rv == ERROR_WANT_READ) {
+                    # Can do! Wait for the next event.
+
+                    # Resume our write if needed.
+                    if (delete $self->{tls_want_write}) {
+                        # Try our write again!
+                        $self->{async_io}->register_write($self->{fileno});
+                    }
+                } elsif ($rv == ERROR_NONE) {
+                    # Huh?
+                } else {
+                    my $error= Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error());
+                    $shutdown_when_done= "TLS error: $error";
+                }
+            }
+
+        } else {
+            my $read_cnt= sysread($self->{socket}, $BUFFER, 10240, $bufsize);
+            if ($read_cnt) {
+                $bufsize += $read_cnt;
+                $read_something= 1;
+
+            } elsif (!defined $read_cnt) {
+                if (!$!{EAGAIN}) {
+                    my $error= "$!";
+                    $shutdown_when_done= $error;
+                }
+            } elsif ($read_cnt == 0) { # EOF
+                $shutdown_when_done= "Disconnected from server";
             }
-        } elsif ($read_cnt == 0) { # EOF
-            $shutdown_when_done= "Disconnected from server";
         }
 
 READ_NEXT:
@@ -811,7 +916,7 @@ READ_MORE:
     }
 
     if ($shutdown_when_done) {
-        $self->shutdown(undef, $shutdown_when_done);
+        $self->shutdown($shutdown_when_done);
     }
 
     return;
@@ -820,21 +925,52 @@ READ_MORE:
 sub can_write {
     my ($self)= @_;
 
-    my $result= syswrite($self->{socket}, $self->{pending_write});
-    if (!defined($result)) {
-        if ($!{EAGAIN}) {
-            return; # Huh. Oh well, whatever
+    if ($self->{tls}) {
+        my $rv= Net::SSLeay::write(${$self->{tls}}, $self->{pending_write});
+        if ($rv > 0) {
+            substr($self->{pending_write}, 0, $rv, '');
+            if (!length $self->{pending_write}) {
+                $self->{async_io}->unregister_write($self->{fileno});
+                delete $self->{pending_write};
+            }
+            return;
+
+        } else {
+            $rv= Net::SSLeay::get_error(${$self->{tls}}, $rv);
+            if ($rv == ERROR_WANT_WRITE) {
+                # Wait until the next callback.
+                return;
+            } elsif ($rv == ERROR_WANT_READ) {
+                # Unschedule ourselves
+                $self->{async_io}->unregister_write($self->{fileno});
+                $self->{tls_want_write}= 1;
+                return;
+            } elsif ($rv == ERROR_NONE) {
+                # Huh?
+                return;
+            } else {
+                my $error= Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error());
+                return $self->shutdown("TLS error: $error");
+            }
         }
 
-        my $error= "$!";
-        return $self->shutdown(undef, $error);
-    }
-    if ($result == 0) { return; } # No idea whether that happens, but guard anyway.
-    substr($self->{pending_write}, 0, $result, '');
+    } else {
+        my $result= syswrite($self->{socket}, $self->{pending_write});
+        if (!defined($result)) {
+            if ($!{EAGAIN}) {
+                return; # Huh. Oh well, whatever
+            }
 
-    if (!length $self->{pending_write}) {
-        $self->{async_io}->unregister_write($self->{fileno});
-        delete $self->{pending_write};
+            my $error= "$!";
+            return $self->shutdown($error);
+        }
+        if ($result == 0) { return; } # No idea whether that happens, but guard anyway.
+        substr($self->{pending_write}, 0, $result, '');
+
+        if (!length $self->{pending_write}) {
+            $self->{async_io}->unregister_write($self->{fileno});
+            delete $self->{pending_write};
+        }
     }
 
     return;
@@ -844,18 +980,18 @@ sub can_timeout {
     my ($self, $id)= @_;
     my $stream= delete $self->{pending_streams}{$id};
     $self->{pending_streams}{$id}= [ sub{}, \(my $zero= 0) ]; # fake it
-    $stream->[0]->(Cassandra::Client::Error->new(message => "Request timed out", is_timeout => 1));
+    $stream->[0]->(Cassandra::Client::Error->new(
+        message         => "Request timed out",
+        is_timeout      => 1,
+        request_error   => 1,
+    ));
     return;
 }
 
 sub shutdown {
-    my ($self, $cb, $shutdown_reason)= @_;
-
-    if ($self->{shutdown}) {
-        $cb->() if $cb;
-        return;
-    }
+    my ($self, $shutdown_reason)= @_;
 
+    return if $self->{shutdown};
     $self->{shutdown}= 1;
 
     my $pending= $self->{pending_streams};
@@ -871,12 +1007,11 @@ sub shutdown {
 
     for (values %$pending) {
         $_->[0]->(Cassandra::Client::Error->new(
-            message   => "Disconnected: $shutdown_reason",
+            message       => "Disconnected: $shutdown_reason",
+            request_error => 1,
         ));
     }
 
-    $cb->() if $cb;
-
     return;
 }
 
