@@ -3,6 +3,8 @@ use 5.010;
 use strict;
 use warnings;
 
+use Encode;
+
 require Exporter;
 our @ISA= qw(Exporter);
 
@@ -251,6 +253,10 @@ sub unpack_stringlist {
 
 # TYPE: bytes
 sub pack_bytes {
+    if (utf8::is_utf8($_[0])) {
+        warn 'BUG: utf8 data passed to pack_bytes';
+        Encode::_utf8_off($_[0]);
+    }
     defined $_[0] ? (pack_int(length($_[0])).$_[0]) : pack_int(-1)
 }
 
@@ -269,6 +275,10 @@ sub unpack_bytes {
 
 # TYPE: shortbytes
 sub pack_shortbytes {
+    if (utf8::is_utf8($_[0])) {
+        warn 'BUG: utf8 data passed to pack_shortbytes';
+        Encode::_utf8_off($_[0]);
+    }
     defined $_[0] ? (pack_short(length($_[0])).$_[0]) : pack_short(-1)
 }
 
